//----------------------------------------------------------------------------

//General control module

//--------------------- Description of the algorithm: ------------------------

//The TControl class provides general control. In the constructor, the equipment
//is initialized and all the necessary objects are created. The Execute method
//is called in the main loop, it polls the button and encoder, as well as
//processes the menu pages. In addition, a number of services are performed
//here: watchdog and saving data to EEPROM. Saving is performed after a power
//failure. The processor is powered through a Schottky diode from a 220 uF
//capacitor, which is enough to save. The built-in comparator monitors the power
//supply. When the power drops below the set threshold, the processor sets
//the output pins to a logic zero state (to prevent current leakage into
//de-energized external circuits) and begins the process of saving data.
//Then he waits for power to be restored. If this occurs, the program continues
//normal execution. If the power continues to drop, the processor will reset.

//----------------------------------------------------------------------------

#include "main.hpp"
#include "control.hpp"
#include "count.hpp"
#include "eeprom.hpp"
#include "setup.hpp"

//------------------------------ Variables: ----------------------------------

TDisplay *Display;
TCount *Count;
TEeprom *Eeprom;
TSetup *Setup;
TEncoder *Encoder;

//----------------------------------------------------------------------------
//---------------------------- TControl class: -------------------------------
//----------------------------------------------------------------------------

//----------------------------- Constructor: ---------------------------------

TControl::TControl(void)
{
  HardwareInit();
  Eeprom = new TEeprom();
  Display = new TDisplay();
  Keyboard = new TKeyboard();
  Encoder = new TEncoder();
  Count = new TCount();
  Menu = new TMenuItems(MENUS);
  MenuTimer = new TSoftTimer(MENU_BRIEF);
  MenuTimer->Oneshot = 1;
  ExitTimer = new TSoftTimer(MENU_EXIT);
  ExitTimer->Oneshot = 1;
  Setup = new TSetup();

  //if the button is held down when turned on,
  if(Keyboard->Key())
  {
    //then output the FW version number:
    Display->Version();
    //waiting for button release:
    do { TSysTimer::Delay_ms(10); }
    while(Keyboard->Key());
  }
  else
  {
    //else check display segments:
    Display->Test();
  }
  TSysTimer::Delay_ms(1000);
  Display->Clear();

  //reading current mode from EEPROM:
  MnuIndex = Eeprom->Rd8Chk(EE_MODE, MNU_DIAMETER, MNU_TIME);
  //select the desired menu:
  Menu->SelectMenu(MnuIndex);
  MenuTimer->Start();
  ExitTimer->Start();
  MenuBrief = 1;
  BackMenu = MNU_TIME;
}

//-------------------------- Control execution: ------------------------------

void TControl::Execute(void)
{
  Display->Execute();
  Keyboard->Execute();
  Encoder->Execute();
  Count->Execute();

  //menu timer processing:
  if(MenuTimer->Over())
  {
    Menu->SelectedMenu->Show();
    MenuBrief = 0;
  }
  //updating the display in the menu:
  if(!MenuBrief)
  {
    Menu->SelectedMenu->Update();
  }
  //read button message:
  KeyMsg_t KeyMsg = Keyboard->Message;
  //read encoder message:
  if(KeyMsg == KEY_NO)
  {
    KeyMsg = Encoder->Message;
  }
  //processing button or encoder message:
  if(KeyMsg != KEY_NO)
  {
    Menu->SelectedMenu->OnKeyboard(KeyMsg);
    //move to another menu if required:
    uint8_t new_menu = Menu->SelectedMenu->MnuIndex;
    if(new_menu != MnuIndex)
    {
      MnuIndex = new_menu;
      Menu->SelectMenu(MnuIndex);
      MenuTimer->Start();
      MenuBrief = 1;
    }
    ExitTimer->Start();
  }
  //exit from the Setup menu:
  if((MnuIndex == MNU_SETUP) && (ExitTimer->Over()))
  {
    Setup->Esc();
    MnuIndex = MNU_TIME;
    Menu->SelectMenu(MnuIndex);
    MenuTimer->Start();
    MenuBrief = 1;
  }

  //services:
  WatchdogService();
  DataSaveService();

  //clear button message:
  Keyboard->Message = KEY_NO;
  //clear encoder message:
  Encoder->Message = KEY_NO;
}

//----------------------------------------------------------------------------
//----------------------- Services Implementation: ---------------------------
//----------------------------------------------------------------------------

//------------------------ Watchdog timer service: ---------------------------

inline void TControl::WatchdogService(void)
{
  if(TSysTimer::Tick)
  {
    __watchdog_reset();
  }
}

//------------------------- Data saving service: -----------------------------

inline void TControl::DataSaveService(void)
{
  if(ACSR & (1 << ACO))
  {
    Display->OnOff(0); //turn off the display to reduce consumption
    Count->EESave();   //saving parameters in EEPROM
    Eeprom->Wr8(EE_MODE, MnuIndex); //saving the current mode
    Eeprom->Validate();       //signature recording
    TSysTimer::Delay_ms(300); //delay
    while(ACSR & (1 << ACO))  //an endless loop if the power is not normal
      __watchdog_reset();
    Display->OnOff(1); //turning on the display if power is restored
  }
}

//----------------------------------------------------------------------------
//---------------------------- Auxiliary Methods: ----------------------------
//----------------------------------------------------------------------------

inline void TControl::HardwareInit(void)
{
  //pins setup:
  Pin_VCC.DirIn(PIN_FLOAT);
  //setting up unused pins:
  Pin_PB4.DirIn(PIN_PULL);
  Pin_PC5.DirIn(PIN_PULL);
  //Pin_RXD.DirIn(PIN_PULL);
  //Pin_TXD.DirIn(PIN_PULL);
  //Pin_DIR.DirIn(PIN_PULL);
  Pin_PD5.DirIn(PIN_PULL);
  Pin_PD6.DirIn(PIN_PULL);
  //watchdog timer setting:
  __watchdog_reset();
  WDTCSR |= (1 << WDCE) | (1 << WDE); //watchdog timer enable
  WDTCSR = (1 << WDE) | (1 << WDP2) | (1 << WDP0); //interval 500 ms
  //analog comparator setting:
  DIDR1 = (1 << AIN1D); //turn off digital buffer on AIN1
  ACSR = (1 << ACBG);   //AIN0 = bandgap reference voltage
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
