//----------------------------------------------------------------------------

//Count implementation module

//--------------------- Description of the algorithm: ------------------------

//The optical sensor consists of two reflective optocouplers placed under
//the chopper with a certain number of BLADES. At the output of the sensor,
//quadrature signals F1 and F2 are formed, which are fed to the ports PB0 and PB1.
//Sensor signals are processed in the Opto_Handler pin change interrupt.
//In the interrupt, based on the analysis of the current and previous states
//of the signals F1 and F2, a decision is made to increment or decrement the
//counter Count. There are 4 increment or decrement events per chopper blade,
//i.e. implemented 4X-quadrature decoder. There are
//EV_PER_REV = BLADES * DECODER = 5 * 4 = 20 events per roller revolution.

//The maximum interrupts frequency for the maximum tape speed during rewind
//Vmax = 10 m/s, roller Diameter = 37.9 mm and number of BLADES = 5 is
//Fmax = Vmax / (Diameter * PI) * BLADES * DECODER =
//10000 mm/s / (37.9 mm * 3.14159) * 5 * 4 = 1680 Hz

//The Count counter is only 8-bit, its task is not to overflow between polls
//by the main program, the period of which is set by the PollTimer software timer.
//The reading of the Count counter is performed when interrupts are disabled,
//after reading it is reset to zero. In fact, this counter contains the
//incremental number of Delta events between two polls. The main program counts
//optical sensor events in the 32-bit Events variable based on the Delta value.

//Based on the Events variable, the metrage value is calculated:
//Metrage = Events * Diameter * PI / EV_PER_REV.
//Calculations are carried out in integer mathematics, to prevent error,
//the Metrage value is calculated in micrometers. The roller diameter value
//is not a constant, but is calculated during the calibration procedure.

//Based on the Metrage variable, the real time value for the selected tape
//speed is calculated: Time = Metrage / Speed.
//The tape speed is selected with a 9/19 logic signal. In the menu, you can
//select speeds 19/38.

//Because the counter is multi-speed, it must overflow at one point for any speed.
//The speeds differ exactly by a factor of 2, so the maximum time for a speed
//of 19.05 cm/s will be 2 times less than for a speed of 9.525 cm/s, for a speed
//of 38.1 cm/s - 4 times less. The time for each speed is calculated from the
//metrage, so it is enough to limit the metrage.

//The time is displayed with a sign in increments of 1 sec or 0.1 sec.
//The maximum time value for each of the speeds is 8.00.00.0, 4.00.00.0 and
//2.00.00.0. The maximum value of the metrage is 2743.1.

//Implemented two independent counters - TimeCounter and LapTimeCounter.

//To calculate the actual tape speed, a meter based on hardware timer 1 is used.
//The meter uses a capture interrupt - one of the optical sensor signals is
//applied to the ICP pin. When the meter starts, the software timer starts with
//the interval M_PRE_TM. If during this time no impulse comes from the sensor
//output, the speed is considered zero. The minimum measurable speed with these
//constant values is approximately 5 cm/s. When a pulse arrives from the optical
//sensor, the timer value is captured into the CapBeg variable, the measuring
//interval M_RUN_TM is started, and the counting of pulses begins. When the
//interval ends, the pulses are counted up to an integer number of roller
//revolutions (otherwise, an error in the manufacture of the chopper will lead
//to a measurement error). Time M_FIN_TM is allocated for the completion of
//the calculation. If during this time the required number of pulses does not
//arrive, the speed is considered zero. At the moment of arrival of the last
//pulse, the counter value is captured and the duration of one revolution
//of the roller is calculated: TurnTime = T_timer * (CapEnd - CapBeg) / Turns
//Then, based on this duration and the diameter of the roller, the speed
//is calculated: Speed = Diameter * PI / TurnTime;

//The tape movement signal is output to the Pin_MOVE port (PD2, XP2 pin 4),
//which can be used instead of the standard tape movement sensor signal.

//The roller diameter calibration procedure uses tape speed meter data.
//In order not to include the old roller diameter (which may have a large error)
//in the calculations, the measured speed value is not used.
//Instead, the turnover duration is used. To improve the accuracy of the
//calibration, the measurement is carried out CAL_CYCLES times, and the total
//time TimeSum is calculated. Then the diameter of the roller is calculated:
//D = (TimeSum * SPEED_19) / (PI * CAL_CYCLES / T_timer)
//With these values of the constants, the duration of the calibration is
//approximately 14 seconds, the measurement is carried out 10 times, each of
//them lasts 2 complete revolutions of the roller.
//If the calculated value of the roller diameter is not within the range of
//30 to 70 mm, then a calibration error occurs.

//----------------------------------------------------------------------------

#include "main.hpp"
#include "count.hpp"
#include "eeprom.hpp"
#include "setup.hpp"

//------------------------------- Constants: ---------------------------------

#ifdef TANDBERG
  static uint8_t  const BLADES = 2;       //blades count for Tandberg
#else
  static uint8_t  const BLADES = 5;       //blades count
#endif
static uint8_t  const DECODER = 4;        //quadrature decoder 4X
static int64_t  const EV_PER_REV = BLADES * DECODER; //events per revolution
static uint16_t const POLL_TM = 10;       //count update period, ms
static uint16_t const PRESCALER = 1024;   //timer 1 prescaler
static int64_t  const PI = 3.14159 * METRAGE_UNIT; //PI * METRAGE_UNIT
enum { SC_9, SC_19, SC_38 };              //speed codes
static int32_t const MIN_MOVE_SPEED = 47; //mm/s
static uint16_t const L_PER_EVENT = PI * NOM_DIA / EV_PER_REV / METRAGE_UNIT; //um
static uint8_t const N_POLL_MOVE = L_PER_EVENT / MIN_MOVE_SPEED / POLL_TM;

//----------------------------------------------------------------------------
//----------------------------- TCount class: --------------------------------
//----------------------------------------------------------------------------

//----------------------------- Constructor: ---------------------------------

TCount::TCount(void)
{
  Pin_F1.DirIn(PIN_PULL);
  Pin_F2.DirIn(PIN_PULL);
  Pin_MOVE.DirOut(PIN_LOW);
  Pin_9_19.DirIn(PIN_PULL);
  Pin_19_38.DirIn(PIN_PULL); //V223
  Pin_EN.DirIn(PIN_PULL);    //V223
  Pin_JMP3.DirIn(PIN_PULL);

  //pin change interrupt enable:
  PCMSK0 = (1 << PCINT1) | (1 << PCINT0);
  PCIFR = (1 << PCIF0);
  PCICR = (1 << PCIE0);

  PollTimer = new TSoftTimer(POLL_TM);
  PollTimer->Autoreload = 1;
  PollTimer->Start();
  TimeCounter = new TUniCounter();
  LapTimeCounter = new TUniCounter();
  Meter = new TMeter();
  NPoll = N_POLL_MOVE;
  Move = 0;
  Run = 0;


  EERead(); //read EEPROM
  UpdateTm = 1;
  UpdateSp = 0;
  CalStatus = CAL_NO;
  SpdSel = OP_H19L9;
  SpeedCode = SC_9;
}

//----------------------- Optical sensor interrupt: --------------------------

volatile int8_t TCount::Count;

#pragma vector = PCINT0_vect
__interrupt void Opto_Handler(void)
{
  static bool preF1 = 0;
  bool nowF1 = TCount::Pin_F1;
  bool nowF2 = TCount::Pin_F2;
  (preF1 == nowF2)? TCount::Count-- : TCount::Count++;
  preF1 = nowF1;
}

//----------------------------- Read count: ----------------------------------

__monitor int8_t TCount::GetCount(void)
{
  int8_t c = Count;
  Count = 0;
  return(c);
}

//--------------------------- Execution count: -------------------------------

void TCount::Execute(void)
{
  //event count:
  if(PollTimer->Over())
  {
    //counter update:
    int8_t Delta = GetCount();
#ifdef EN_9_19_38
    if(Delta && Pin_EN)
#else
    if(Delta)
#endif
    {
      TimeCounter->AddDelta(Delta);
      LapTimeCounter->AddDelta(Delta);
      UpdateTm = 1;
      //Move signal generation:
      Run = Delta;
    }
    if(!--NPoll)
    {
      NPoll = N_POLL_MOVE;
      Move = Run;
      Pin_MOVE = Move; //Move output
      Run = 0;
    }
    //speed selection:
    uint8_t sc = SC_19;

#ifdef EN_9_19_38
    if(Pin_19_38)
    {
      if(Pin_9_19) sc = SC_19;
        else sc = SC_9;
    }
    else
    {
      if(Pin_9_19) sc = SC_38;
        else sc = SC_19;
    }
#else
    switch(SpdSel)
    {
    case OP_H19L9:  sc = Pin_9_19? SC_19 : SC_9;  break;
    case OP_H9L19:  sc = Pin_9_19? SC_9  : SC_19; break;
    case OP_H38L19: sc = Pin_9_19? SC_38 : SC_19; break;
    case OP_H19L38: sc = Pin_9_19? SC_19 : SC_38; break;
    }
#endif

    if(SpeedCode != sc)
    {
      SpeedCode = sc;
      TimeCounter->SetSpeed(sc);
      LapTimeCounter->SetSpeed(sc);
      UpdateTm = 1;
    }
  }
  //tape speed measurement:
  Meter->Execute(Move);
  if(Meter->Update())
  {
    UpdateSp = 1;
    //roller diameter calibration:
    if(CalStatus == CAL_RUN)
    {
      TimeSum += Meter->GetTurnTime();
      if(++CalCnt == CAL_CYCLES)
      {
        if(CalculateDiameter())
          CalStatus = CAL_OK;
            else CalStatus = CAL_ERR;
      }
    }
  }
}

//--------------------------- Read count update: -----------------------------

bool TCount::UpdateTime(void)
{
  bool upd = UpdateTm;
  UpdateTm = 0;
  return(upd);
}

//--------------------------- Read speed update: -----------------------------

bool TCount::UpdateSpeed(void)
{
  bool upd = UpdateSp;
  UpdateSp = 0;
  return(upd);
}

//------------------ Start roller diameter calibration: ----------------------

void TCount::CalibStart(uint8_t s)
{
  TimeSum = 0;
  CalCnt = 0;
  CalStatus = s;
}

//--------------------- Roller diameter calculation: -------------------------

bool TCount::CalculateDiameter(void)
{
  int64_t const C1 = PI * F_CLK * CAL_CYCLES / PRESCALER / METRAGE_UNIT;
  int64_t const C2 = C1 / 2;
  uint32_t d = ((int64_t)TimeSum * SPEED_19 + C2) / C1;
  if(d < MIN_DIA || d > MAX_DIA) return(0);
  Diameter = d;
  TimeCounter->SetDiameter(Diameter);
  LapTimeCounter->SetDiameter(Diameter);
  Meter->SetDiameter(Diameter);
  Eeprom->Wr32(EE_DIAMETER, Diameter);
  UpdateTm = 1;
  return(1);
}

//----------------- Setting the nominal roller diameter: ---------------------

void TCount::SetNomDiameter(void)
{
  Diameter = NOM_DIA;
  TimeCounter->SetDiameter(Diameter);
  LapTimeCounter->SetDiameter(Diameter);
  Meter->SetDiameter(Diameter);
  Eeprom->Wr32(EE_DIAMETER, Diameter);
}


//-------------------- Reading parameters from EEPROM: -----------------------

void TCount::EERead(void)
{
  Diameter = Eeprom->Rd32(EE_DIAMETER, NOM_DIA);
  TimeCounter->SetDiameter(Diameter);
  LapTimeCounter->SetDiameter(Diameter);
  Meter->SetDiameter(Diameter);
  TimeCounter->EERead(EE_TIME);
  LapTimeCounter->EERead(EE_LAPTIME);
}

//-------------------- Writing parameters to EEPROM: -------------------------

void TCount::EESave(void)
{
  Eeprom->Wr32(EE_DIAMETER, Diameter);
  TimeCounter->EESave(EE_TIME);
  LapTimeCounter->EESave(EE_LAPTIME);
}

//----------------------------------------------------------------------------
//--------------------------- TUniCounter calss: -----------------------------
//----------------------------------------------------------------------------

//----------------------------- Constructor: ---------------------------------

TUniCounter::TUniCounter(void)
{
  Events = 0;
  Speed = SPEED_9;
}

//--------------------------- Events counting: -------------------------------

void TUniCounter::AddDelta(int8_t d)
{
  Events = Events + d;
  Metrage = Events2Metrage(Events);
  //signed, tenths of a second:
  if(Metrage >= MAX_UM_S10)
  {
    Events = Metrage2Events(Metrage - MAX_UM_S10 + MIN_UM_S10) - 1;
    Metrage = Events2Metrage(Events);
  }
  else if(Metrage <= MIN_UM_S10)
  {
    Events = Metrage2Events(Metrage - MIN_UM_S10 + MAX_UM_S10) + 1;
    Metrage = Events2Metrage(Events);
  }
  Time = Metrage2Time(Metrage);
}

//----------------------- Calculation of metrage: ----------------------------

int64_t TUniCounter::Events2Metrage(int32_t e)
{
  return(((int64_t)e * Diameter * PI / EV_PER_REV) / METRAGE_UNIT);
}

//------------------ Calculating the number of events: -----------------------

int32_t TUniCounter::Metrage2Events(int64_t m)
{
  return(m * METRAGE_UNIT / ((int64_t)Diameter * PI / EV_PER_REV));
}

//------------------------- Time calculation: --------------------------------

int32_t TUniCounter::Metrage2Time(int64_t m)
{
  return(m / Speed);
}

//-------------------------- Reading metrage: --------------------------------

int32_t TUniCounter::GetMetrage(void)
{
  return((Metrage) / (METRAGE_UNIT / 10));
}

//---------------------------- Reading time: ---------------------------------

int32_t TUniCounter::GetTime(void)
{
  return(Time);
}

//--------------------------- Reset counter: ---------------------------------

void TUniCounter::Reset(void)
{
  Events = 0;
  Metrage = 0;
  Time = 0;
}

//---------------------- Setting the current speed: --------------------------

void TUniCounter::SetSpeed(int8_t s)
{
  if(s == SC_9) Speed = SPEED_9;
    else if(s == SC_19) Speed = SPEED_19;
      else Speed = SPEED_38;
  Time = Metrage2Time(Metrage);
}

//---------------------- Roller diameter setting: ----------------------------

void TUniCounter::SetDiameter(uint32_t d)
{
  Diameter = d;
  Metrage = Events2Metrage(Events);
  Time = Metrage2Time(Metrage);
}

//------------------- Reading parameters from EEPROM: ------------------------

void TUniCounter::EERead(int8_t a)
{
  Events = Eeprom->Rd32(a);
  Metrage = Events2Metrage(Events);
  Time = Metrage2Time(Metrage);
}

//-------------------- Saving parameters in EEPROM: --------------------------

void TUniCounter::EESave(int8_t a)
{
  Eeprom->Wr32(a, Events);
}

//----------------------------------------------------------------------------
//---------------------------- TMeter class: ---------------------------------
//----------------------------------------------------------------------------

//----------------------------- Constructor: ---------------------------------

TMeter::TMeter(void)
{
  TCCR1B = (1 << ICNC1) | (1 << CS12) | (1 << CS10); //filter, CK/1024
  TIFR1 = (1 << ICF1);
  TIMSK1 |= (1 << ICIE1); //enable interrupt on capture
  IntervalTimer = new TSoftTimer();
  IntervalTimer->Oneshot = 1;
  Speed = 0;
  State = ST_STOP;
}

//------------------------- Interrupt on capture: ----------------------------

uint8_t volatile TMeter::State;
uint16_t TMeter::CapBeg;
uint16_t TMeter::CapEnd;
uint8_t  TMeter::Blade;
uint8_t  TMeter::Turns;

#pragma vector = TIMER1_CAPT_vect
__nested __interrupt void Capture_Handler(void)
{
  switch(TMeter::State)
  {
  case TMeter::ST_START:            //start measurement
    {
      TMeter::CapBeg = ICR1;        //initial timer capture
      TMeter::Turns = 0;            //number of roller revolutions
      TMeter::Blade = 0;            //chopper blade number
      TMeter::State = TMeter::ST_PRE;
      break;
    }
  case TMeter::ST_PRE:              //beginning of measurement
  case TMeter::ST_RUN:              //measurement is running
    {
      if(++TMeter::Blade == BLADES) //blade count
      {
        TMeter::Blade = 0;
        TMeter::Turns++;            //counting whole revolutions
      }
      break;
    }
  case TMeter::ST_FINISH:           //end of measurement
    {
      if(++TMeter::Blade == BLADES) //blade count
      {
        TMeter::CapEnd = ICR1;      //end timer capture
        TMeter::Turns++;            //counting whole revolutions
        TMeter::State = TMeter::ST_END;
      }
      break;
    }
  }
}

//---------------------- Speed measurement process: --------------------------

void TMeter::Execute(int8_t move)
{
  switch(State)
  {
  case ST_STOP:
    IntervalTimer->Start(M_PRE_TM);
    State = ST_START;
    break;
  case ST_PRE:
    IntervalTimer->Start(M_RUN_TM);
    State = ST_RUN;
    break;
  case ST_RUN:
    if(IntervalTimer->Over())
    {
      IntervalTimer->Start(M_FIN_TM);
      State = ST_FINISH;
    }
    break;
  case ST_START:
  case ST_FINISH:
    if(IntervalTimer->Over())
    {
      Speed = 0;
      UpdateSp = 1;
      State = ST_STOP;
    }
    break;
  case ST_END:
    Speed = CalcSpeed();
    UpdateSp = 1;
    State = ST_STOP;
    break;
  }
  if(!move && Speed)
  {
    Speed = 0;
    UpdateSp = 1;
    State = ST_STOP;
  }
}

//--------------------------- Speed calculation: -----------------------------

uint32_t TMeter::CalcSpeed(void)
{
  TurnTime = (uint32_t)(CapEnd - CapBeg) * 10 / Turns;
  int32_t s = Diameter * (PI * F_CLK * 10 / PRESCALER / METRAGE_UNIT) / TurnTime;
  return((s + 5) / 10);
}

//----------------------- Roller diameter setting: ---------------------------

void TMeter::SetDiameter(uint32_t d)
{
  Diameter = d;
}

//----------------------------- Update check: --------------------------------

bool TMeter::Update(void)
{
  bool upd = UpdateSp;
  UpdateSp = 0;
  return(upd);
}

//------------------ Reading the duration of the revolution: -----------------

uint32_t TMeter::GetTurnTime(void)
{
  return(TurnTime);
}

//----------------------------- Reading speed: -------------------------------

uint32_t TMeter::GetSpeed(void)
{
  return(Speed);
}

//----------------------------------------------------------------------------
