//----------------------------------------------------------------------------

//Count implementation module, header file

//----------------------------------------------------------------------------

#ifndef Count_HPP
#define Count_HPP

//------------------------------- Constants: ---------------------------------

static uint8_t const CAL_CYCLES = 10; //number of cycles during calibration
enum cstatus_t { CAL_NO, CAL_RUN, CAL_OK, CAL_ERR }; //calibration status

static int16_t const SEC_PER_MIN = 60;
static int16_t const MIN_PER_HRS = 60;
static int32_t const SEC_PER_HRS = SEC_PER_MIN * MIN_PER_HRS;

static uint32_t const SPEED_9  =  9525; //speed 1, resolution 0.001 cm/s
static uint32_t const SPEED_19 = 19050; //speed 2, resolution 0.001 cm/s
static uint32_t const SPEED_38 = 38100; //speed 3, resolution 0.001 cm/s

static uint32_t const MAX_SPEED = 999999; //max. speed, x 0.001 cm/s

static int64_t const METRAGE_UNIT = 100000; //x 1 um = 0.1 m
static uint32_t const SPEED_UNIT = 10; //x 1 um/s = 0.001 cm/s

//roller diameter:
#ifdef REVOX_A700
static int32_t const MAX_DIA = 70000; //x 0.001 mm
static int32_t const NOM_DIA = 37900; //x 0.001 mm
static int32_t const MIN_DIA = 30000; //x 0.001 mm
#endif
#ifdef REVOX_B77
static int32_t const MAX_DIA = 25000; //x 0.001 mm Revox B77
static int32_t const NOM_DIA = 19000; //x 0.001 mm Revox B77
static int32_t const MIN_DIA = 12000; //x 0.001 mm Revox B77
#endif
#ifdef TANDBERG
static int32_t const MAX_DIA = 16000; //x 0.001 mm Tandberg
static int32_t const NOM_DIA = 12000; //x 0.001 mm Tandberg
static int32_t const MIN_DIA =  9000; //x 0.001 mm Tandberg
#endif

//number ranges (signed, tenths of a second):
static int64_t const MAX_TIME_S10 = 8 * SEC_PER_HRS;
static int64_t const MIN_TIME_S10 = -MAX_TIME_S10;
static int64_t const MAX_UM_S10 = MAX_TIME_S10 * SPEED_9 * SPEED_UNIT;
static int64_t const MIN_UM_S10 = MIN_TIME_S10 * SPEED_9 * SPEED_UNIT;

//----------------------------------------------------------------------------
//-------------------------- TUniCounter class: ------------------------------
//----------------------------------------------------------------------------

class TUniCounter
{
private:
  uint32_t Diameter;
  int32_t Events;
  int64_t Metrage;
  int32_t Time;
  int32_t Speed;
  int64_t Events2Metrage(int32_t e);
  int32_t Metrage2Events(int64_t m);
  int32_t Metrage2Time(int64_t m);

public:
  TUniCounter(void);
  void SetSpeed(int8_t s);
  void SetDiameter(uint32_t d); //resolution 0.001 mm
  void AddDelta(int8_t d);
  void Reset(void);
  int32_t GetMetrage(void);     //resolution 0.1 m / 0.01 m
  int32_t GetTime(void);        //resolution 1 s
  void EERead(int8_t a);
  void EESave(int8_t a);
};

//----------------------------------------------------------------------------
//---------------------------- TMeter class: ---------------------------------
//----------------------------------------------------------------------------

#pragma vector = TIMER1_CAPT_vect
extern "C" __interrupt void Capture_Handler(void);

class TMeter
{
private:
  friend __interrupt void Capture_Handler(void);
  enum { ST_STOP, ST_START, ST_PRE, ST_RUN, ST_FINISH, ST_END };
  static uint16_t const M_PRE_TM =  500; //measurement waiting interval, ms
  static uint16_t const M_RUN_TM = 1000; //measurement interval, ms
  static uint16_t const M_FIN_TM = 2000; //measurement completion interval, ms
  static volatile uint8_t State;
  static uint16_t CapBeg;
  static uint16_t CapEnd;
  static uint8_t Blade;
  static uint8_t Turns;
  TSoftTimer *IntervalTimer;
  uint32_t Diameter;
  uint32_t TurnTime;
  uint32_t Speed;
  bool UpdateSp;
  uint32_t CalcSpeed(void);
public:
  TMeter(void);
  void Execute(int8_t move);
  void SetDiameter(uint32_t d); //resolution 0.001 mm
  bool Update(void);
  uint32_t GetTurnTime(void);   //resolution 6.4 us
  uint32_t GetSpeed(void);      //resolution 0.001 cm/s
};

//----------------------------------------------------------------------------
//---------------------------- TCount class: ---------------------------------
//----------------------------------------------------------------------------

#pragma vector = PCINT0_vect
extern "C" __interrupt void Opto_Handler(void);

class TCount
{
private:
  friend __interrupt void Opto_Handler(void);
  static Pin_F2_t Pin_F2;
  static Pin_F1_t Pin_F1;
  static volatile int8_t Count;
  Pin_MOVE_t Pin_MOVE;
  Pin_9_19_t Pin_9_19;
  Pin_19_38_t Pin_19_38; //V223
  Pin_EN_t Pin_EN;       //V223
  Pin_JMP3_t Pin_JMP3;
  int8_t GetCount(void);
  TSoftTimer *PollTimer;
  TUniCounter *TimeCounter;
  TUniCounter *LapTimeCounter;
  TMeter *Meter;
  bool UpdateTm;
  bool UpdateSp;
  int8_t NPoll;
  int8_t Run;
  int8_t Move;
  int32_t Diameter;
  uint8_t CalStatus;
  uint8_t CalCnt;
  int32_t TimeSum;
  uint8_t SpeedCode;
  uint8_t SpdSel;

public:
  TCount(void);
  void Execute(void);
  bool UpdateTime(void);
  int32_t GetTime(void) { return(TimeCounter->GetTime()); };
  int32_t GetMetrage(void) { return(TimeCounter->GetMetrage()); };
  void ResetTime(void) { TimeCounter->Reset(); UpdateTm = 1; };
  int32_t GetLapTime(void) { return(LapTimeCounter->GetTime()); };
  int32_t GetLapMetrage(void) { return(LapTimeCounter->GetMetrage()); };
  void ResetLapTime(void) { LapTimeCounter->Reset(); UpdateTm = 1; };
  bool UpdateSpeed(void);
  int8_t TapeMove(void) { return(Move); };
  int32_t GetSpeed(void) { return(Meter->GetSpeed()); };
  int32_t GetDiameter(void) { return(Diameter); };
  void CalibStart(uint8_t s);
  uint8_t CalibStatus(void) { return(CalStatus); };
  bool CalculateDiameter(void);
  void SetNomDiameter(void);
  void SetSpdSel(uint8_t s) { SpdSel = s; };
  void EERead(void);
  void EESave(void);
};

//----------------------------------------------------------------------------

extern TCount *Count;

//----------------------------------------------------------------------------

#endif

