
//----------------------------------------------------------------------------

//LED display support module

//-------------------------- Resources used: ---------------------------------

//A 7-segment LED display is used. The indication is static, the indicators
//are connected to the outputs of an external shift register (chain from 74HC595).
//The registers are loaded using the hardware SPI port.
//To access the register, the TSreg class is implemented.
//Ports used: SCLK - PB5 (SPI SCK), SDATA - PB3 (SPI MOSI),
//LOAD - PC0, OE - PC1.

//Time output is implemented in formats "-9.59.59", "99.59.59", "9.59.59.9".
//Metrage output is implemented in formats "-9999.9", "99999.9".
//Speed output is implemented in format "999.999".
//Roller diameter output is implemented in format "d 99.99".
//The diameter display mode supports the output of text messages
//"d CAL" and "d Err", which are used in the calibration mode.
//Any digit on the display can be switched to blinking mode.

//----------------------------------------------------------------------------

#include "main.hpp"
#include "display.hpp"
#include "count.hpp"
#include "setup.hpp"

//------------------------------- Constants: ---------------------------------

//arrangement of segments on the indicator:
//
//    -- A --
//   |       |
//   F       B
//   |       |
//    -- G --
//   |       |
//   E       C
//   |       |
//    -- D --  (H)

//indicator segments in the order they are connected to the outputs
//of the registers on the printed circuit board:
enum Segs_t
{ SEG_A, SEG_B, SEG_C, SEG_H, SEG_D, SEG_E, SEG_F, SEG_G };

//segment masks:
enum SegMask_t
{
  A = (1 << SEG_A),
  B = (1 << SEG_B),
  C = (1 << SEG_C),
  D = (1 << SEG_D),
  E = (1 << SEG_E),
  F = (1 << SEG_F),
  G = (1 << SEG_G),
  H = (1 << SEG_H)
};

//----------------------------------------------------------------------------
//----------------------------- TSreg class: ---------------------------------
//----------------------------------------------------------------------------

//--------------------------- Initialization: --------------------------------

void TSreg::Init(bool on)
{
  Pin_SS.DirOut(PIN_LOW);
  if(on)
  {
    Pin_OE.DirOut(PIN_HIGH);
    Pin_LOAD.DirOut(PIN_HIGH);
    Pin_SDATA.DirOut(PIN_LOW);
    Pin_SCLK.DirOut(PIN_HIGH);
    //SPI setup:
    SPCR = (1 << SPE) | (1 << MSTR) | (1 << CPOL)  | (1 << CPHA);
  }
  else
  {
    //SPI disable:
    SPCR = 0;
    Pin_OE.DirIn(PIN_FLOAT);
    Pin_LOAD.DirIn(PIN_FLOAT);
    Pin_SDATA.DirIn(PIN_FLOAT);
    Pin_SCLK.DirIn(PIN_FLOAT);
  }
}

//------------------------- Write byte via SPI: ------------------------------

void TSreg::SpiWr(uint8_t d)
{
#ifdef COMMONANODE
  SPDR = ~d; //common anode
#else
  SPDR = d; //common cathode
#endif
  while(!(SPSR & (1 << SPIF)));
}

//--------------------------- Loading registers: -----------------------------

void TSreg::Load(uint8_t *Value)
{
  Pin_LOAD = 0;
  for(uint8_t n = 0; n < REGS; n++)
    SpiWr(Value[n]);
  Pin_LOAD = 1;
  Pin_OE = 0;
}

//----------------------------------------------------------------------------
//---------------------------- TDisplay class: -------------------------------
//----------------------------------------------------------------------------

//--------------- Convert character code to 7-segment code: ------------------

uint8_t TDisplay::Seven(uint8_t d)
{
  //font table:
  static const uint8_t Font[] =
  {                            //code - symbol
    A | B | C | D | E | F    , //  0  -  "0"
    B | C                    , //  1  -  "1"
    A | B | D | E | G        , //  2  -  "2"
    A | B | C | D | G        , //  3  -  "3"
    B | C | F | G            , //  4  -  "4"
    A | C | D | F | G        , //  5  -  "5"
    A | C | D | E | F | G    , //  6  -  "6"
    A | B | C                , //  7  -  "7"
    A | B | C | D | E | F | G, //  8  -  "8"
    A | B | C | D | F | G    , //  9  -  "9"
    0                        , // 10  -  BLANK
    G                        , // 11  -  MINUS

    C | D | E | F | G        , // 12  -  ALT6
    A | B | C | F            , // 13  -  ALT7
    A | B | C | F | G        , // 14  -  ALT9

    A | B | C | E | F | G    , // 15  -  "A"
    A | D | E | F            , // 16  -  ""
    D | E | G                , // 17  -  "c"
    B | C | D | E | G        , // 18  -  "d"
    A | D | E | F | G        , // 19  -  "E"
    A | E | F | G            , // 20  -  "F"
    B | C | E | F | G        , // 21  -  "H"
    D | E | F                , // 22  -  "L"
    C | E | G                , // 23  -  "n"
    A | B | E | F | G        , // 24  -  "P"
    E | G                    , // 25  -  "r"
    D | E | F | G            , // 26  -  "t"
    B | C | D | E | F          // 27  -  "U"
  };

  //storing and deleting a point:
  bool p = d & POINT;
  d &= ~POINT;
  //transcoding letters and special characters:
  if(d > ALT9)
  {
    //digit character encoding:
    if(d >= '0' && d <= '9')
    {
      d = d - '0';
    }
    else
    {
      switch(d)
      {
      case ' ': d = BLANK; break;
      case '-': d = MINUS; break;
      case 'A': d = 15; break;
      case 'C': d = 16; break;
      case 'c': d = 17; break;
      case 'd': d = 18; break;
      case 'E': d = 19; break;
      case 'I': d =  1; break;
      case 'F': d = 20; break;
      case 'H': d = 21; break;
      case 'L': d = 22; break;
      case 'n': d = 23; break;
      case 'O': d =  0; break;
      case 'P': d = 24; break;
      case 'r': d = 25; break;
      case 'S': d =  5; break;
      case 't': d = 26; break;
      case 'U': d = 27; break;
      default:  d = MINUS;
      }
    }
  }
  //apply an alternative font for the numbers 6, 7, 9:
  if((d == 6) && (Fnt679 & OP_7AA)) d = 12;
  if((d == 7) && (Fnt679 & OP_A69)) d = 13;
  if((d == 9) && (Fnt679 & OP_7AA)) d = 14;
  //7-segment coding:
  char c = Font[d];
  //point recovery:
  if(p) c |= H;
  return(c);
}

//------------------ Displaying a digit on the indicator: --------------------

void TDisplay::Digit(uint8_t d)
{
  if(d == HOME) Pos = 0;
    else if(d == '.') Segs[Pos - 1] |= H;
      else Segs[Pos++] = Seven(d);
}

//---------------- Filling the indicator with a given symbol: ----------------

void TDisplay::Digits(uint8_t d)
{
  Digit(HOME);
  for(uint8_t i = 0; i < DIGS; i++)
    Digit(d);
}

//------------------------ Output string from Flash: -------------------------

void TDisplay::String(uint8_t __flash *s)
{
  while(*s) Digit(*s++);
}

#define StringF(x) {static __flash uint8_t str[] = {x}; String(str);}

//-------------------- Converting a number to BCD format: --------------------

//Convert Binary to BCD:
//v - input binary number (32 bits, unsigned)
//buff - output BCD array (6 or 7 digits)

void TDisplay::Bcd(uint32_t v, uint8_t *buff)
{
  for(uint8_t p = 0; p < DIGS; p++)
    buff[p] = 0;                      //clearing the output buffer
  for(uint8_t i = 0; i < 32; i++)     //cycle by number of input bits
  {
    uint8_t c = (v >> 31) & 1;        //save carry
    v = v << 1;                       //input number shift
    for(uint8_t p = 0; p < DIGS; p++) //cycle by number of digits
    {
      uint8_t s = buff[p];            //reading digit
      s = (s << 1) | c; c = 0;        //shift with carry
      if(s > 9) { s += 0x06; c = 1; } //digit correction
      s &= 0x0F;                      //nibble extraction
      buff[p] = s;                    //save digit
    }
  }
}

//----------------------------------------------------------------------------
//---------------------------- Public methods: -------------------------------
//----------------------------------------------------------------------------

//----------------------------- Constructor: ---------------------------------

TDisplay::TDisplay(void)
{
  BlinkTimer = new TSoftTimer(BLINK_TM / 2);
  BlinkTimer->Autoreload = 1;
  BlinkEn = 0;
  BlinkOn = 0;
  Fnt679 = 0;
  TimeHR = 1;
  SpeedHR = 1;
  Sreg.Init(1);
  Clear();
}

//----------------------------- Blink execute: -------------------------------

void TDisplay::Execute(void)
{
  if(BlinkEn && TSysTimer::Tick)
  {
    if(BlinkTimer->Over())
    {
      BlinkOn = !BlinkOn;
      Sreg.Load(BlinkOn? Segs : Darks);
    }
  }
}

//----------------------------- Blink enable: --------------------------------

void TDisplay::Blink(uint8_t p)
{
  if(p)
  {
    for(int8_t i = 0; i < DIGS; i++)
      Darks[i] = (p & (1 << i))? 0 : Segs[i];
  }
  BlinkEn = p;
  BlinkOn = 1;
  BlinkTimer->Start();
  Sreg.Load(Segs);
}

//------------------ Font selection for numbers 6, 7, 9: ---------------------

void TDisplay::SetFont(uint8_t fnt)
{
  Fnt679 = fnt;
}

//------------------- Time display resolution selection: ---------------------

void TDisplay::SetTimeRes(uint8_t op)
{
  TimeHR = op == OP_RTCHI;
}

//------------------ Length display resolution selection: --------------------

void TDisplay::SetMetrageRes(uint8_t op)
{
  MetrageHR = op == OP_LENHI;
}

//------------------- Speed display resolution selection: --------------------

void TDisplay::SetSpeedRes(uint8_t op)
{
  SpeedHR = (op == OP_SPDHI) || (op == OP_SPDHL);
  SpeedHL = op == OP_SPDHL;
}

//---------------------------- Clear display: --------------------------------

void TDisplay::Clear(void)
{
  Digits(BLANK);
  BlinkEn = 0;
  Sreg.Load(Segs);
}

//----------------------------- Display test: --------------------------------

void TDisplay::Test(void)
{
  Digits(8 + POINT);
  BlinkEn = 0;
  Sreg.Load(Segs);
}

//---------------------------- Display on/off: -------------------------------

void TDisplay::OnOff(bool on)
{
  Sreg.Init(on);
  if(on) Sreg.Load(Segs);
}

//---------------- Outputting text from the first position: ------------------

void TDisplay::Text(uint8_t __flash *s)
{
  Digit(HOME);
  String(s);
  Sreg.Load(Segs);
}

//------------------ Time output -H.MM.SS.D or -H.MM.SS: ---------------------

void TDisplay::Time(int32_t v, bool lap)
{
  Digit(HOME);
  //sign store:
  bool minus = v < 0;
  if(minus) v = -v;
  //separation of tenths of a second:
  uint8_t sec10 = v % 10; v = v / 10;
  //rounding if tenth seconds are not displayed:
  if(!TimeHR && (sec10 >= 5)) v = v + 1;
  //sec -> hrs.min.sec conversion:
  uint8_t hrs = v / SEC_PER_HRS;
  uint16_t m = v % SEC_PER_HRS;
  uint8_t min = m / SEC_PER_MIN;
  uint8_t sec = m % SEC_PER_MIN;
  //output to the indicator:
  Digit(minus? MINUS : BLANK);
  Digit(hrs);
  Digit('.');
  Digit(min / 10);
  Digit(min % 10);
  Digit('.');
  Digit(sec / 10);
  Digit(sec % 10);
  if(TimeHR)
  {
    Digit('.');
    Digit(sec10);
    if(lap) Digit('.');
  }
  else
  {
    if(lap) Digit('.');
    Digit(BLANK);
  }
  BlinkEn = 0;
  Sreg.Load(Segs);
}

//-------------- Metrage output in -MMMM.MM or -MMMM.M format: ---------------

void TDisplay::Metrage(int32_t v, bool lap)
{
  //Leading zeros are not suppressed.
  Digit(HOME);
  bool minus = v < 0;
  if(minus) v = -v;
  //rounding if hundredths of meters are not displayed:
  if(!MetrageHR && (v % 10 >= 5)) v = v + 10;
  //conversion to BCD:
  Bcd(v, Digs);
  Digit(minus? MINUS : BLANK);
  Digit(Digs[5]);
  Digit(Digs[4]);
  Digit(Digs[3]);
  Digit(Digs[2]);
  Digit('.');
  Digit(Digs[1]);
  if(MetrageHR)
  {
    Digit(Digs[0]);
    if(lap) Digit('.');
  }
  else
  {
    if(lap) Digit('.');
    Digit(BLANK);
  }
  BlinkEn = 0;
  Sreg.Load(Segs);
}

//--------------- Speed output in SSS.SSS or SSS.SS format: ------------------

void TDisplay::Speed(uint32_t v)
{
  Digit(HOME);
  if(v > MAX_SPEED)
  {
    //overflow display:
    StringF("   -.---");
  }
  else
  {
    //rounding if thousandths of cm/s are not displayed:
    if(!SpeedHR && (v % 10 >= 5)) v = v + 10;
    //conversion to BCD:
    Bcd(v, Digs);
    //output to the indicator:
    if(!SpeedHL) Digit(BLANK);
    if(!Digs[5]) Digs[5] = BLANK;
    Digit(Digs[5]);
    if(!Digs[4] && Digs[5] == BLANK) Digs[4] = BLANK;
    Digit(Digs[4]);
    Digit(Digs[3]);
    Digit('.');
    Digit(Digs[2]);
    Digit(Digs[1]);
    Digit(SpeedHR? Digs[0] : BLANK);
  }
  BlinkEn = 0;
  Sreg.Load(Segs);
}

//--------------- Roller diameter output in d DD.DD format: ------------------

void TDisplay::Diameter(uint32_t v, uint8_t s)
{
  Digit(HOME);
  StringF("d ");
  if(s == D_CAL)
  {
    //calibration string output:
    StringF(" CAL ");
  }
  else if(s == D_ERR)
  {
    //error message output:
    StringF(" Err ");
  }
  else
  {
    //conversion to BCD:
    Bcd(v, Digs);
    //displaying the value on the indicator:
    Digit(Digs[4]);
    Digit(Digs[3]);
    Digit('.');
    Digit(Digs[2]);
    Digit(Digs[1]);
    Digit(Digs[0]);
  }
  BlinkEn = 0;
  Sreg.Load(Segs);
}

//--------------- Display FW version number in V.VV format: ------------------

void TDisplay::Version(void)
{
  Digit(HOME);
  Bcd(VERSION, Digs);
  Digit(BLANK);
  Digit(BLANK);
  Digit(Digs[2]);
  Digit('.');
  Digit(Digs[1]);
  Digit(Digs[0]);
  Digit(BLANK);
  Digit(BLANK);
  BlinkEn = 0;
  Sreg.Load(Segs);
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
