//----------------------------------------------------------------------------

//Menu implementation module

//--------------------- Description of the algorithm: ------------------------

//The counter is controlled using the menu system. For control, a single button
//or a button plus an encoder is used. The button uses short press,
//hold for 1 second and long hold for 4 seconds.

//By turning the encoder knob, you can select the display modes.

//The count values can be reset by briefly pressing the encoder button.
//"rtc-1" and "LEn-1" are different representations of the same count value
//and are therefore reset at the same time. The same is true for the "rtc-2"
//and "LEn-2" pair.

//Resetting is not possible for speed readings, so pressing the encoder button
//will display the roller diameter. If you now press and hold the encoder button
//for 4 seconds, the roller diameter calibration mode will start.

//When calibration starts, "d CAL" flashes on the display. Calibration lasts
//about 14 seconds, at the end the display will show a new value for the roller
//diameter. If the calibration fails (for example, the tape did not move or its
//speed was very different from the nominal), the display will show "d Err".
//In this case, you can either repeat the calibration by long holding the button,
//or exit it by short pressing, while the new value will not be saved.
//If the encoder button is pressed during calibration, the process will be
//interrupted. If the button is held for 1 second, the default value of
//the roller diameter will be loaded. After calibration, the encoder button must
//be pressed to accept the new roller diameter value.

//When you hold the encoder button for 1 second from any display mode, you enter
//the configuration menu. In this case, "SEtUP" appears on the display,
//and after 1 sec. the menu is displayed. The transition between menu items is
//done by turning the encoder, the choice of options for each item is made by
//a series of short presses.

//----------------------------------------------------------------------------

#include "main.hpp"
#include "menu.hpp"
#include "control.hpp"
#include "count.hpp"
#include "setup.hpp"

//----------------------------------------------------------------------------
//--------------------------- TMenuTime class: -------------------------------
//----------------------------------------------------------------------------

//-------------------------- Display menu name: ------------------------------

void TMenuTime::Brief(void)
{
  Display_TextF(" rtc-1 ");
}

//---------------------------- Display value: --------------------------------

void TMenuTime::Show(void)
{
  Display->Time(Count->GetTime());
}

//----------------------------- Update value: --------------------------------

void TMenuTime::Update(void)
{
  if(Count->UpdateTime()) Show();
}

//---------------------------- Keyboard event: -------------------------------

void TMenuTime::OnKeyboard(KeyMsg_t &msg)
{
#ifdef ENCODER
  if(msg == KEY_REL)  Count->ResetTime();
  if(msg == KEY_UP)   MnuIndex = Setup->NextMenu(MNU_TIME);
  if(msg == KEY_DN)   MnuIndex = Setup->PrevMenu(MNU_TIME);
  if(msg == KEY_HOLD) MnuIndex = MNU_SETUP;
#else
  if(msg == KEY_HOLD) Count->ResetTime();
  if(msg == KEY_REL)  MnuIndex = Setup->NextMenu(MNU_TIME);
  if(msg == KEY_LONG) MnuIndex = MNU_SETUP;
#endif
}

//----------------------------------------------------------------------------
//-------------------------- TMenuMetrage class: -----------------------------
//----------------------------------------------------------------------------

//-------------------------- Display menu name: ------------------------------

void TMenuMetrage::Brief(void)
{
  Display_TextF(" LEn-1 ");
}

//---------------------------- Display value: --------------------------------

void TMenuMetrage::Show(void)
{
  Display->Metrage(Count->GetMetrage());
}

//----------------------------- Update value: --------------------------------

void TMenuMetrage::Update(void)
{
  if(Count->UpdateTime()) Show();
}

//---------------------------- Keyboard event: -------------------------------

void TMenuMetrage::OnKeyboard(KeyMsg_t &msg)
{
#ifdef ENCODER
  if(msg == KEY_REL)  Count->ResetTime();
  if(msg == KEY_UP)   MnuIndex = Setup->NextMenu(MNU_METRAGE);
  if(msg == KEY_DN)   MnuIndex = Setup->PrevMenu(MNU_METRAGE);
  if(msg == KEY_HOLD) MnuIndex = MNU_SETUP;
#else
  if(msg == KEY_HOLD) Count->ResetTime();
  if(msg == KEY_REL)  MnuIndex = Setup->NextMenu(MNU_METRAGE);
  if(msg == KEY_LONG) MnuIndex = MNU_SETUP;
#endif
}

//----------------------------------------------------------------------------
//-------------------------- TMenuLapTime class: -----------------------------
//----------------------------------------------------------------------------

//-------------------------- Display menu name: ------------------------------

void TMenuLapTime::Brief(void)
{
  Display_TextF(" rtc-2 ");
}

//---------------------------- Display value: --------------------------------

void TMenuLapTime::Show(void)
{
  Display->Time(Count->GetLapTime(), TDisplay::LAP);
}

//----------------------------- Update value: --------------------------------

void TMenuLapTime::Update(void)
{
  if(Count->UpdateTime()) Show();
}

//---------------------------- Keyboard event: -------------------------------

void TMenuLapTime::OnKeyboard(KeyMsg_t &msg)
{
#ifdef ENCODER
  if(msg == KEY_REL)  Count->ResetLapTime();
  if(msg == KEY_UP)   MnuIndex = Setup->NextMenu(MNU_LAPTIME);
  if(msg == KEY_DN)   MnuIndex = Setup->PrevMenu(MNU_LAPTIME);
  if(msg == KEY_HOLD) MnuIndex = MNU_SETUP;
#else
  if(msg == KEY_HOLD) Count->ResetLapTime();
  if(msg == KEY_REL)  MnuIndex = Setup->NextMenu(MNU_LAPTIME);
  if(msg == KEY_LONG) MnuIndex = MNU_SETUP;
#endif
}

//----------------------------------------------------------------------------
//------------------------ TMenuLapMetrage class: ----------------------------
//----------------------------------------------------------------------------

//-------------------------- Display menu name: ------------------------------

void TMenuLapMetrage::Brief(void)
{
  Display_TextF(" LEn-2 ");
}

//---------------------------- Display value: --------------------------------

void TMenuLapMetrage::Show(void)
{
  Display->Metrage(Count->GetLapMetrage(), TDisplay::LAP);
}

//----------------------------- Update value: --------------------------------

void TMenuLapMetrage::Update(void)
{
  if(Count->UpdateTime()) Show();
}

//---------------------------- Keyboard event: -------------------------------

void TMenuLapMetrage::OnKeyboard(KeyMsg_t &msg)
{
#ifdef ENCODER
  if(msg == KEY_REL)  Count->ResetLapTime();
  if(msg == KEY_UP)   MnuIndex = Setup->NextMenu(MNU_LAPMETRAGE);
  if(msg == KEY_DN)   MnuIndex = Setup->PrevMenu(MNU_LAPMETRAGE);
  if(msg == KEY_HOLD) MnuIndex = MNU_SETUP;
#else
  if(msg == KEY_HOLD) Count->ResetLapTime();
  if(msg == KEY_REL)  MnuIndex = Setup->NextMenu(MNU_LAPMETRAGE);
  if(msg == KEY_LONG) MnuIndex = MNU_SETUP;
#endif
}

//----------------------------------------------------------------------------
//--------------------------- TMenuSpeed class: ------------------------------
//----------------------------------------------------------------------------

//-------------------------- Display menu name: ------------------------------

void TMenuSpeed::Brief(void)
{
  Display_TextF(" SPEEd ");
}

//---------------------------- Display value: --------------------------------

void TMenuSpeed::Show(void)
{
  Display->Speed(Count->GetSpeed());
}

//----------------------------- Update value: --------------------------------

void TMenuSpeed::Update(void)
{
  if(Count->UpdateSpeed()) Show();
}

//---------------------------- Keyboard event: -------------------------------

void TMenuSpeed::OnKeyboard(KeyMsg_t &msg)
{
#ifdef ENCODER
  if(msg == KEY_REL)  MnuIndex = MNU_DIAMETER;
  if(msg == KEY_UP)   MnuIndex = Setup->NextMenu(MNU_SPEED);
  if(msg == KEY_DN)   MnuIndex = Setup->PrevMenu(MNU_SPEED);
  if(msg == KEY_HOLD) MnuIndex = MNU_SETUP;
#else
  if(msg == KEY_HOLD) MnuIndex = MNU_DIAMETER;
  if(msg == KEY_REL)  MnuIndex = Setup->NextMenu(MNU_SPEED);
  if(msg == KEY_LONG) MnuIndex = MNU_SETUP;
#endif
}

//----------------------------------------------------------------------------
//-------------------------- TMenuDiameter class: ----------------------------
//----------------------------------------------------------------------------

//-------------------------- Display menu name: ------------------------------

void TMenuDiameter::Brief(void)
{
  Display->Diameter(Count->GetDiameter());
}

//---------------------------- Display value: --------------------------------

void TMenuDiameter::Show(void)
{
  Update();
}

//----------------------------- Update value: --------------------------------

void TMenuDiameter::Update(void)
{
  uint8_t status = Count->CalibStatus();
  if(status == CAL_OK)
  {
    Display->Diameter(Count->GetDiameter());
    Count->CalibStart(CAL_NO);
  }
  else if(status == CAL_ERR)
  {
    Display->Diameter(0, TDisplay::D_ERR);
    Count->CalibStart(CAL_NO);
  }
}

//---------------------------- Keyboard event: -------------------------------

void TMenuDiameter::OnKeyboard(KeyMsg_t &msg)
{
  //button hold:
  if(msg == KEY_HOLD)
  {
    //start calibration:
    if(Count->CalibStatus() != CAL_RUN)
    {
      Display->Diameter(0, TDisplay::D_CAL);
      Display->Blink(TDisplay::BLINK);
      Count->CalibStart(CAL_RUN);
    }
    else
    {
      //when calibration is running - default diameter:
      Count->CalibStart(CAL_NO);
      Count->SetNomDiameter();
      Display->Diameter(Count->GetDiameter());
    }
  }
  //short press - exit calibration:
  if(msg == KEY_REL)
  {
    if(Count->CalibStatus() == CAL_RUN)
    {
      Count->CalibStart(CAL_NO);
      Display->Diameter(Count->GetDiameter());
    }
    else
    {
      Count->CalibStart(CAL_NO);
      MnuIndex = MNU_SPEED;
    }
  }
}

//----------------------------------------------------------------------------
//--------------------------- TMenuSetup class: ------------------------------
//----------------------------------------------------------------------------

//-------------------------- Display menu name: ------------------------------

void TMenuSetup::Brief(void)
{
  Display_TextF(" SEtUP ");
  SubMenu = SM_ESC;
}

//---------------------------- Display value: --------------------------------

void TMenuSetup::Show(void)
{
  Setup->Show(SubMenu);
}

//----------------------------- Update value: --------------------------------

void TMenuSetup::Update(void)
{
}

//---------------------------- Keyboard event: -------------------------------

void TMenuSetup::OnKeyboard(KeyMsg_t &msg)
{
#ifdef ENCODER
  //turn right - increase the menu index:
  if(msg == KEY_UP)
  {
    if(SubMenu < SM_MAX)
    {
      SubMenu++;
      Show();
    }
  }
  //turn left - decrease the menu index:
  if(msg == KEY_DN)
  {
    if(SubMenu > SM_MIN)
    {
      SubMenu--;
      Show();
    }
  }
#else
  //press - next menu index:
  if(msg == KEY_REL)
  {
    if(SubMenu < SM_MAX)
      SubMenu++;
        else SubMenu = SM_MIN;
    Show();
  }
#endif
#ifdef ENCODER
  //press - change options:
  if(msg == KEY_REL)
#else
  //button hold - change options:
  if(msg == KEY_HOLD)
#endif
  {
    switch(SubMenu)
    {
    case SM_ESC: Setup->Esc(); MnuIndex = MNU_TIME; break;
    case SM_DEF: Setup->Def(); MnuIndex = MNU_TIME; break;
    case SM_SET: Setup->Set(); MnuIndex = MNU_TIME; break;
    default: Setup->Edit(SubMenu); Show();
    }
  }
}

//----------------------------------------------------------------------------
//---------------------------- TMenuItems class: -----------------------------
//----------------------------------------------------------------------------

//----------------------------- Constructor: ---------------------------------

TMenuItems::TMenuItems(char max) : TList(max)
{
  AddItem(new TMenuTime());
  AddItem(new TMenuMetrage());
  AddItem(new TMenuLapTime());
  AddItem(new TMenuLapMetrage());
  AddItem(new TMenuSpeed());
  AddItem(new TMenuDiameter());
  AddItem(new TMenuSetup());
}

//------------------------- Go to menu by index: -----------------------------

void TMenuItems::SelectMenu(uint8_t mnu)
{
  SelectedMenu = Items[mnu];
  SelectedMenu->MnuIndex = mnu;
  SelectedMenu->Brief();
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
