//----------------------------------------------------------------------------

//System timer module

//----------------------------------------------------------------------------

#include "main.hpp"
#include "systimer.hpp"

//---------------------------- Resources used: -------------------------------

//The system timer is implemented on the basis of a timer 0 interrupt with
//a period of 1 ms. A system tick with this period is distributed to all
//processes. In addition, 16-bit software timers (TSoftTimer) work in this
//time base, which can be created as needed.

//----------------------------------------------------------------------------
//--------------------------- TSysTimer class: -------------------------------
//----------------------------------------------------------------------------

//--------------------- System timer initializing: ---------------------------

void TSysTimer::Init(void)
{
  TCCR0A = (1 << WGM01);              //CTC
  TCCR0B = (1 << CS01) | (1 << CS00); //CK/64
  OCR0A = TC0_VAL - 1;                //period loading
  TIFR0 = (1 << OCF0A);               //clearing pending interrupts
  TIMSK0 = (1 << OCIE0A);             //compare interrupt enable
  vTick = 1;
  vCounter = 0;
  Counter = 0;
  __enable_interrupt();               //interrupts enable
}

//------------------------ System timer interrupt: ---------------------------

volatile bool TSysTimer::vTick;
volatile uint16_t TSysTimer::vCounter;

#pragma vector = TIMER0_COMPA_vect
__nested __interrupt void SysTick_Handler(void)
{
  TSysTimer::vCounter++;
  TSysTimer::vTick = 1;
}

//------------------- Synchronization of system ticks: -----------------------

bool TSysTimer::Tick;
uint16_t TSysTimer::Counter;

__monitor void TSysTimer::Sync(void)
{
  Counter = vCounter;
  Tick = vTick;
  vTick = 0;
}

//------------------------ Atomic counter reading: ---------------------------

__monitor uint16_t TSysTimer::GetCount(void)
{
  return(vCounter);
}

//------------------- Millisecond range delay function: ----------------------

void TSysTimer::Delay_ms(uint16_t d)
{
  uint16_t DelayStart = GetCount();
  do { Sync(); __watchdog_reset(); }
    while(Counter - DelayStart < d);
}

//---------------------- Timeout millisecond range: --------------------------

uint16_t TSysTimer::Start_ms;
uint16_t TSysTimer::Interval_ms;

void TSysTimer::TimeoutStart_ms(uint16_t t)
{
  Start_ms = GetCount();
  Interval_ms = t;
}

bool TSysTimer::TimeoutOver_ms(void)
{
  return(GetCount() - Start_ms >= Interval_ms);
}

//----------------------------------------------------------------------------
//-------------------------- TSoftTimer class: -------------------------------
//----------------------------------------------------------------------------

//---------------------------- Constructor: ----------------------------------

TSoftTimer::TSoftTimer(uint16_t t)
{
  Interval = t;
  Autoreload = 0;
  Oneshot = 0;
  fRun = 0;
  fEvent = 0;
  StartCount = TSysTimer::Counter;
}

//-------------------------------- Start: ------------------------------------

void TSoftTimer::Start(uint16_t t)
{
  Interval = t;
  fRun = 1;
  fEvent = 0;
  StartCount = TSysTimer::Counter;
}

void TSoftTimer::Start()
{
  fRun = 1;
  fEvent = 0;
  StartCount = TSysTimer::Counter;
}

//--------------------------------- Stop: ------------------------------------

void TSoftTimer::Stop()
{
  fRun = 0;
}

//--------------------------- Interval setting: ------------------------------

void TSoftTimer::SetInterval_ms(uint16_t t)
{
  Interval = t;
}

void TSoftTimer::SetInterval_sec(uint16_t t)
{
  Interval = t * 1000;
}

//---------------------------- Forced overflow: ------------------------------

void TSoftTimer::Force(void)
{
  fRun = 1;
  StartCount = TSysTimer::Counter - Interval;
}

//--------------------------- Reading overflow: ------------------------------

bool TSoftTimer::Over(void)
{
  bool fevent = TSysTimer::Counter - StartCount >= Interval;
  if(fevent)
  {
    if(Oneshot && fEvent) fevent = 0;
    fEvent = 1;
    if(Autoreload) StartCount = TSysTimer::Counter;
  }
  if(Oneshot || Autoreload) return(fevent && fRun);
  return(fEvent && fRun);
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
